
#include "main.h"
#include "struct_3d.h"
#include "camera.h"
#include "render.h"
#include "keybind.h"
#include "input.h"
#include "3dmode.h"
#include "tex_browser.h"
#include "editor_window.h"
#include "3d_window.h"
#include "map.h"

#include <wx/image.h>

//wxCursor* blank_cursor = NULL;

CVAR(Bool, invert_mouse_3d, false, CVAR_SAVE)
CVAR(Int, vid_width_3d, -1, CVAR_SAVE)
CVAR(Int, vid_height_3d, -1, CVAR_SAVE)
CVAR(Int, vid_bpp_3d, -1, CVAR_SAVE)
CVAR(Int, vid_refresh_3d, -1, CVAR_SAVE)

extern wxGLContext *gl_context;
extern EditorWindow *editor_window;
extern Camera camera;
extern BindList binds;


EXTERN_CVAR(Float, mouse_speed_3d)

#ifdef WIN32
DEVMODE current_mode;
#endif

void set_3d_resolution()
{
#ifdef WIN32
	EnumDisplaySettings(0, ENUM_CURRENT_SETTINGS, &current_mode);

	if (vid_width_3d != -1 &&
		vid_height_3d != -1 &&
		vid_bpp_3d != -1 &&
		vid_refresh_3d != -1)
	{
		DEVMODE new_mode;
		new_mode.dmFields = DM_PELSWIDTH|DM_PELSHEIGHT|DM_BITSPERPEL|DM_DISPLAYFREQUENCY;
		new_mode.dmPelsWidth = vid_width_3d;
		new_mode.dmPelsHeight = vid_height_3d;
		new_mode.dmBitsPerPel = vid_bpp_3d;
		new_mode.dmDisplayFrequency = vid_refresh_3d;

		long ret = ChangeDisplaySettings(&new_mode, CDS_FULLSCREEN);

		if (ret == DISP_CHANGE_RESTART)
			wxMessageBox(_T("Restart required to set screen resolution"));

		if (ret == DISP_CHANGE_BADFLAGS)
			wxMessageBox(_T("Error: Bad flags"));

		if (ret == DISP_CHANGE_BADPARAM)
			wxMessageBox(_T("Error: Bad parameters"));

		if (ret == DISP_CHANGE_FAILED)
			wxMessageBox(_T("Resolution change failed"));

		if (ret == DISP_CHANGE_BADMODE)
			wxMessageBox(_T("Resolution not supported"));

		//if (ret == DISP_CHANGE_NOTUPDATED)
		//	wxMessageBox(_T("Restart required to set screen resolution!"));
	}
#endif
}

void restore_resolution()
{
#ifdef WIN32
	ChangeDisplaySettings(&current_mode, 0);
#endif
}

Render3dCanvas::Render3dCanvas(wxWindow *parent)
:	wxGLCanvas(parent, gl_context, -1)
{
}

Render3dCanvas::~Render3dCanvas()
{
	editor_window->Show(true);
	restore_resolution();
}

void Render3dCanvas::render()
{
	SetCurrent();

	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	render_3d_view(GetClientSize().x, GetClientSize().y);

	SwapBuffers();
}

BEGIN_EVENT_TABLE(Render3dCanvas, wxGLCanvas)
	EVT_IDLE(Render3dCanvas::idle_event)
	EVT_PAINT(Render3dCanvas::paint_event)
	EVT_KEY_DOWN(Render3dCanvas::key_down)
	EVT_KEY_UP(Render3dCanvas::key_up)
	EVT_MOUSE_EVENTS(Render3dCanvas::mouse_event)
	EVT_SIZE(Render3dCanvas::size_event)
END_EVENT_TABLE()

void Render3dCanvas::do_loop()
{
	if (wxGetElapsedTime(false) >= 10)
	{
		// Reset timer
		wxStartTimer();

		// Process input
		if (!keys_3d())
		{
			if (GetParent())
				GetParent()->Destroy();

			return;
		}

		// Gravity
		if (camera.gravity)
			apply_gravity();

		// Get hilighted item
		determine_hilight();
	}

	// Render
	render();
}

void Render3dCanvas::idle_event(wxIdleEvent &event)
{
	do_loop();

	event.RequestMore();
}

void Render3dCanvas::paint_event(wxPaintEvent &event)
{
	wxPaintDC dc(this);
	render();
}

void Render3dCanvas::key_down(wxKeyEvent &event)
{
	binds.set(get_key_name(event.GetKeyCode()), NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());
}

void Render3dCanvas::key_up(wxKeyEvent &event)
{
	binds.unset(get_key_name(event.GetKeyCode()), NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());
}

bool reset = true;
void Render3dCanvas::mouse_event(wxMouseEvent &event)
{
	if (event.LeftDown() || event.LeftDClick())
		binds.set("Mouse1", NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());

	if (event.MiddleDown() || event.MiddleDClick())
		binds.set("Mouse2", NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());

	if (event.RightDown() || event.RightDClick())
		binds.set("Mouse3", NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());

	if (event.LeftUp())
		binds.unset("Mouse1", NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());

	if (event.MiddleUp())
		binds.unset("Mouse2", NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());

	if (event.RightUp())
		binds.unset("Mouse3", NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());

	if (event.GetWheelRotation() > 0)
	{
		binds.set("MWheel Up", NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());
		keys_3d(true);
	}
	else if (event.GetWheelRotation() < 0)
	{
		binds.set("MWheel Down", NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());
		keys_3d(true);
	}

	binds.unset("MWheel Up", NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());
	binds.unset("MWheel Down", NULL, event.ShiftDown(), event.ControlDown(), event.AltDown());

	if (event.Moving())
	{
		if (reset)
		{
			reset = false;
			return;
		}

		int center_x = GetClientSize().x * 0.5;
		int center_y = GetClientSize().y * 0.5;
		float angle_x = -(event.GetX() - center_x) * (0.001f * mouse_speed_3d);
		float angle_y = -(event.GetY() - center_y) * (0.001f * mouse_speed_3d);

		if (invert_mouse_3d)
			angle_y = -angle_y;

		point3_t axis = cross(camera.view - camera.position, camera.up_vector);
		axis = axis.normalize();

		camera.rotate_view(angle_x, 0, 0, 1);
		camera.rotate_view(angle_y, axis.x, axis.y, axis.z);

		reset = true;
		WarpPointer(center_x, center_y);
	}
}

void Render3dCanvas::size_event(wxSizeEvent &event)
{
	SetCurrent();

	GLfloat fogColor[4]= {0.0f, 0.0f, 0.0f, 1.0f};

	glViewport(0, 0, GetClientSize().x, GetClientSize().y);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0);
	glShadeModel(GL_SMOOTH);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glCullFace(GL_FRONT);
	glEnable(GL_CULL_FACE);
	glEnable(GL_TEXTURE_2D);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glDepthRange(0, 1);

	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogfv(GL_FOG_COLOR, fogColor);
	glFogf(GL_FOG_DENSITY, 2.0f);
	glHint(GL_FOG_HINT, GL_NICEST);
	glFogf(GL_FOG_START, 0.0f);
	glFogf(GL_FOG_END, 40.0f);
	glEnable(GL_FOG);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(60.0f, (GLfloat)320/(GLfloat)200, 0.1f, 1000.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	render();
}

void start_3d_mode()
{
	if (map.n_sectors <= 1)
	{
		wxMessageBox(_T("You need at least 2 sectors to start 3d mode"));
		return;
	}

	// Setup
	if (!setup_3d_data())
		return;

	// Set resolution
	set_3d_resolution();

	// Create window
	wxFrame *frame_3d = new wxFrame(NULL, -1, _T("SLADE (3d Mode)"));
	Render3dCanvas *canvas_3d = new Render3dCanvas(frame_3d);
	wxBoxSizer *sizer = new wxBoxSizer(wxHORIZONTAL);
	frame_3d->SetSizer(sizer);
	sizer->Add(canvas_3d, 1, wxEXPAND);

	// Hide cursor
	wxImage img(32, 32, true);
	img.SetMask(true);
	img.SetMaskColour(0, 0, 0);
	frame_3d->SetCursor(wxCursor(img));

	editor_window->Hide();
	frame_3d->ShowFullScreen(true);
	canvas_3d->render();
	wxStartTimer();
}
