
#include "main.h"
#include "editor_window.h"
#include "map.h"
#include "wad_manager.h"
#include "game_config.h"
#include "splash.h"
#include "keybind.h"
#include "colours.h"
#include "console.h"

#include <wx/image.h>
#include <wx/sysopt.h>
#include <wx/tipdlg.h>
#include <wx/tooltip.h>

#ifdef __APPLE__
// Doing our own thing for Apple path handling, since 10.4 only ships
// with wxWidgets 2.5, and besides we need to make sure to write to
// the user's path instead of /Applications
#include <CoreFoundation/CFString.h>
#include <CoreFoundation/CFBundle.h>
#include <cstdlib>
string userPath;
#elif wxCHECK_VERSION(2, 6, 0)
// Everyone else uses wxWidgets-provided paths
#include <wx/stdpaths.h>
#endif

WadManager		*wad_manager;
EditorWindow	*editor_window;

string app_path;

CVAR(Bool, reopen_wads, true, CVAR_SAVE)
CVAR(String, col_config, "Default", CVAR_SAVE)
CVAR(Bool, tips_start, 1, CVAR_SAVE)
CVAR(Int, tip_current, 0, CVAR_SAVE)
CVAR(Bool, tabs_theme, 1, CVAR_SAVE)

extern WadList wads;

extern BindList binds;

string parse_string(char *str, ...)
{
	char text[512] = "";
	va_list ap;

	va_start(ap, str);
	vsprintf(text, str, ap);
	va_end(ap);

	string ret = text;
	return ret;
}

// str_upper: Returns a string in uppercase
// ------------------------------------- >>
string str_upper(string str)
{
	string retval = str;
	for(unsigned i = 0; i < str.length(); ++i)
		str[i] = (char)toupper(str[i]);
	return retval;
}

// wx_to_str: Converts a wxString to a std::string (also from unicode to ascii)
// ------------------------------------------------------------------------- >>
string wx_to_str(wxString str)
{
	string ret;

	wxCharBuffer buffer = str.ToAscii();
	const char* data = buffer;
	ret = data;

	return ret;
}

// str_to_wx: Converts a std::string to a wxString
// -------------------------------------------- >>
wxString str_to_wx(string str)
{
#if wxUSE_UNICODE
	if(!str.length())
		return wxString();

	wchar_t *multibyteString = new wchar_t[str.length() * 5];
	mbstowcs(multibyteString, str.c_str(), str.length() * 5);
	wxString ret(multibyteString);
	delete[] multibyteString;
	return ret;
#else
	return wxString(str.c_str());
#endif
}

// load_main_config: Loads the main SLADE configuration file
// ------------------------------------------------------ >>
void load_main_config(bool wads_open)
{
	Tokenizer mr;

	if (!mr.open_file("slade.cfg", 0, 0))
		return;

	string token = mr.get_token();

	while (token != "!END")
	{
		if (token == "cvars")
			load_cvars(&mr);

		if (token == "iwads")
			load_game_iwads(&mr);

		if (token == "recent_wads")
			load_recent_wads(&mr);

		if (token == "open_wads" && reopen_wads && !wads_open)
			load_open_wads(&mr);

		if (token == "key_binds")
			binds.load(&mr);
		
		/*
		if (token == "window_props")
			load_window_properties(&mr);
			*/

		token = mr.get_token();
	}
}

// save_main_config: Saves the main SLADE configuration file
// ------------------------------------------------------ >>
void save_main_config()
{
#ifdef __APPLE__
	FILE *fp = fopen((userPath + "/slade.cfg").c_str(), "wt");
#else
	FILE *fp = fopen("slade.cfg", "wt");
#endif
	save_cvars(fp);
	save_game_iwads(fp);
	binds.save(fp);
	save_recent_wads(fp);
	if (reopen_wads) save_open_wads(fp);
	//save_window_properties(fp);
	fclose(fp);
}

IMPLEMENT_APP(MainApp)

bool MainApp::OnInit()
{
	srand(wxGetLocalTime());

#ifdef __APPLE__
	CFURLRef bundleRef = CFBundleCopyBundleURL(CFBundleGetMainBundle());
	CFStringRef bundlePath = CFURLCopyFileSystemPath(bundleRef, kCFURLPOSIXPathStyle);
	wxString dataDir = str_to_wx(string(CFStringGetCStringPtr(bundlePath, CFStringGetSystemEncoding())) + "/Contents/Resources/");
	userPath = string(getenv("HOME")) + string("/Library/Application Support/Slade");
	if(!wxDirExists(str_to_wx(userPath)))
		wxMkdir(str_to_wx(userPath), 0700);
#elif wxCHECK_VERSION(2, 6, 0)
	wxStandardPaths sp;
	wxString dataDir = sp.GetDataDir();
#endif

	wxSetWorkingDirectory(dataDir);

#ifdef __APPLE__
	wxLog::SetActiveTarget(new wxLogStderr(fopen((userPath + "/slade.log").c_str(), "wt")));
#else
	wxLog::SetActiveTarget(new wxLogStderr(fopen("slade.log", "wt")));
#endif
	wxSystemOptions::SetOption(wxT("msw.remap"), 0);

	// Disable xp themes on tabs if the option is enabled
	if (!tabs_theme)
		wxSystemOptions::SetOption(wxT("msw.notebook.themed-background"), 0);

	wxImage::AddHandler(new wxPNGHandler);
	wxToolTip::Enable(true);

	app_path = wx_to_str(dataDir);

#ifdef WIN32
	app_path += "\\";
#else
	app_path += "/";
#endif

	wxLogMessage(_T("%s"), app_path.c_str());

	setup_splash();
	splash("Starting Up...", false);

	bool wads_opened = false;
	if (argc > 0)
	{
		for (int a = 0; a < argc; a++)
			wxLogMessage(argv[a]);

		for (int a = 1; a < argc; a++)
		{
			wxString arg = argv[a];

			if (arg.Right(4) == _T(".wad") ||
				arg.Right(4) == _T(".WAD") ||
				arg.Right(4) == _T(".Wad"))
			{
				wxLogMessage(_T("Opening wad %s"), wx_to_str(arg).c_str());
				wads.open_wad(wx_to_str(arg));
				wads_opened = true;
			}
		}
	}

	load_main_config(wads_opened);

	editor_window = new EditorWindow(_T("SLADE"), -1, -1, 800, 600);
	SetTopWindow(editor_window);

	init_console();

	load_colour_configs();
	string ccfg = col_config;
	set_colour_config(ccfg);

	wad_manager = new WadManager(editor_window);
	splash_hide();

	if (tips_start)
    {
		wxTipProvider *tipProvider = wxCreateFileTipProvider(_T("res/tips.txt"), tip_current);
		tips_start = wxShowTip(editor_window, tipProvider);
		tip_current = tipProvider->GetCurrentTip();
		delete tipProvider;
    }

	return true;
}

int MainApp::OnExit()
{
	save_main_config();

	return 0;
}
